@@ -2,6 +2,7 @@ import uuid
 import json
 import time
 import os
+import zipfile
 
 import requests
 import pandas as pd
@@ -115,10 +116,10 @@ class Client(object):
             The process id is used to check the progress
             and to fetch the results after it is finished.
         """
-        pid_resp = self.session.get(
-            self.service_url+'process_id', 
-            data={'process_tag':'DESCRIPTORS'+uuid.uuid4().hex}
-        )
+        #pid_resp = self.session.get(
+        #    self.service_url+'process_id', 
+        #    data={'process_tag':'DESCRIPTORS'+uuid.uuid4().hex}
+        #)
         procid = pid_resp.json()['process_id']
         resp = self.session.post(
             self.service_url+'api/get_descriptors', 
@@ -176,28 +177,24 @@ class Client(object):
             for structures with fractionally occupied sites,
             for 'lithium_ionic' or 'anion_redox' featurization.
         return_files : bool
-            If True, a zip files is returned containing 
-            descriptors as json and all sampled structures as POSCAR files,
+            If True, a zip file is returned containing 
+            descriptors as a .csv file
+            and all sampled structures as POSCAR files,
             for 'lithium_ionic' or 'anion_redox' featurization.
 
         Returns
         -------
         response : dict
-            Dict containing status report and process id. 
-            The process id is used to fetch the results 
+            Dict containing status report and file tag. 
+            The results can be fetched by file tag
+            via client.download_descriptor_results()
             after the process is finished.
         """
-        pid_resp = self.session.get(
-            self.service_url+'process_id', 
-            data={'process_tag':'DESCRIPTORS'+uuid.uuid4().hex}
-        )
-        procid = pid_resp.json()['process_id']
         if featurization in ['lithium_ionic','anion_redox']:
             resp = self.session.post(
                 self.service_url+'api/compute_descriptors', 
                 data={
                     'featurization':featurization,
-                    'process_id':procid,
                     'structure_random_seed':structure_random_seed,
                     'structure_sample_size':structure_sample_size,
                     'return_files':return_files
@@ -209,18 +206,15 @@ class Client(object):
                 self.service_url+'api/compute_descriptors', 
                 data={
                     'featurization':featurization,
-                    'process_id':procid,
                     'source_data':source_data
                 },
             )
-
         if resp.status_code == 200:
             return resp.json()
         else:
             return {'success':False,'status_code':resp.status_code}
 
-    # TODO: unify with download_process_results (see download_apply_model_results)
-    def download_descriptor_results(self,process_id,wait=False,wait_interval=3.,output_path=None):
+    def download_descriptor_results(self,file_tag,wait=False,wait_interval=3.,output_path=None):
         """Get status or results of a descriptor process.
 
         This is used to download the results of a call to
@@ -234,21 +228,21 @@ class Client(object):
 
         Parameters
         ----------
-        process_id : int
-            The process id assigned to the descriptor computation.
-            This id is returned by the client
-            after a call to get_descriptors() or compute_descriptors().
+        file_tag : str 
+            The tag of the output file assigned to the descriptor computation.
+            This is returned as part of the compute_descriptors() response.
         wait : bool
-            Specifies whether or not to wait for the process to finish.
-            If wait==False and the process is not finished,
-            a process status report will be returned.
+            Specifies whether or not to wait for the file to show up. 
+            If wait==False and the file is not present,
+            the response will indicate a missing file.
         wait_interval : float
             If wait==True, this is the time delay (in seconds)
-            between status checks. The interval must be at least one second.
-            If a value less than 1 is provided, it is ignored.
+            between attempts to check for the file. 
+            The interval must be at least one second.
+            If a value less than 1 is provided, it is set to 1.
         output_path : str 
             If output_path is provided, apyonics attempts to save 
-            the table of results to this path.
+            the results to this path.
 
         Returns
         -------
@@ -256,53 +250,39 @@ class Client(object):
             Dict containing either a report of process status or a table of results.
         """
         if wait_interval < 1: wait_interval = 1.
-        # check the process
-        resp = self.session.get(
-            self.service_url+'process_status',
-            params={'process_id':process_id}
-        )
-        if resp.status_code == 200: 
-            procstat = resp.json()
-        else:
-            procstat = {'status':'ERROR','status_code':resp.status_code} 
         n_waits = 0
-        if wait:
-            while not procstat['status'] in ['FINISHED','ERROR']: 
-                time.sleep(wait_interval)
-                procstat = self.session.get(
-                    self.service_url+'process_status',
-                    params={'process_id':process_id}
-                ).json()
+        file_found = False
+        while not file_found and wait: 
+            # check for the file
+            try:
+                res = self.session.get(
+                    self.service_url+'download_file',
+                    json={'file_tag':file_tag}
+                )
+                data = res._content 
+                file_found = bool(data)
+            except Exception as ex:
+                print('{}'.format(ex))
                 n_waits += 1
                 # gradually slow down status checks
                 if np.mod(n_waits,100) == 0 and wait_interval < 30:
                     wait_interval += 1
-        if procstat['status'] == 'FINISHED':
-            # NOTE: if it was a fractionally occupied crystal,
-            # and the job was submitted with return_files==True,
-            # this will return a zip file instead of csv!
-            res = self.session.get(
-                self.service_url+'download_file',
-                json={'file_tag':'DESCRIPTORS'+str(process_id)}
-            )
+                time.sleep(wait_interval)
+        # we now presume the process to be finished, 
+        # and the output file to be present
+        try:
             if output_path:
                 with open(output_path,'wb') as f:
-                    f.write(res._content)
-            data = res._content 
-            data_df = None
-            try:
-                # NOTE: this should fail if the res._content is a zip file
-                data_df = pd.read_csv(BytesIO(res._content),index_col='name')
-            except:
-                pass
-            return {'success':True,'data':data,'df':data_df}
-        else:
-            procstat['success'] = False
-            if procstat['status'] == 'ERROR':
-                procstat['error'] = 'process failed' 
-            else:
-                procstat['error'] = 'process not finished' 
-            return procstat
+                    f.write(data)
+        except:
+            return {'success':False,'error':'unable to download file'}
+        zf = zipfile.ZipFile(BytesIO(data))
+        data_df = None
+        try:
+            data_df = pd.read_csv(zf.open('descriptors.csv'),index_col='name')
+        except:
+            pass
+        return {'success':True,'data_df':data_df,'data':data,'output_path':output_path}
 
 
     # DATASETS
@@ -444,13 +424,14 @@ class Client(object):
         resp = self.session.post(
             self.service_url+'api/new_table',
             files={
-                'table':open(filepath,'rb'),
+                'table':BytesIO(open(filepath,'rb').read()),
                 'metadata':json.dumps({
                     'dataset_id':dataset_id,
                     'name':name,
                     'description':description
                 }).encode()
-            }
+            },
+            verify=False
         )
         if resp.status_code == 200:
             return resp.json()
@@ -990,6 +971,7 @@ class Client(object):
             Each substance can have the following settings:
 
             - 'featurization' (bool): Specifies whether or not the substance should be featurized. 
+            - 'use_descriptors' (bool): Optional, default True. Includes any available descriptors in featurization. 
             - 'descriptor_set' (str): Descriptor subset (optional). See `get_descriptors()`.
             - 'concentration_weighted' (bool): Optional, default False. Multiplies the descriptors by substance concentrations. For a column named 'substance1', the concentration data is assumed to be in a column named 'substance1_conc'.
             - 'in_solution' (bool): Optional, default False. Substances in solution get their descriptors combined additively with the descriptors of other substances in the same solution.
@@ -1023,6 +1005,10 @@ class Client(object):
                 data[subsk+'_solnname'] = stgs['solution_name']
             if 'use_properties' in stgs:
                 data[subsk+'_useprops'] = stgs['use_properties']
+            if 'use_descriptors' in stgs:
+                data[subsk+'_usedescs'] = stgs['use_descriptors']
+            else:
+                data[subsk+'_usedescs'] = True
         resp = self.session.post(
             self.service_url+'featurize', 
             json=data
@@ -1090,14 +1076,16 @@ class Client(object):
 
                 - 'none' (keep all or select manually)
                 - 'variances' (rank inputs in order of decreasing variance)
-                - 'rfe_coefs' (recursive elimination by regression coefficients)
+                - 'rfe_coefs' (recursive elimination by regression coefficients or feature importances)
                 - 'combinatorial' (full combinatorial inspection) 
 
             - 'inspection_model' (str): Options:
 
                 - 'ridge': Ridge regression- only for 'rfe_coefs' strategy
+                - 'rf_regressor': Random forest regression- only for 'rfe_coefs' strategy 
                 - 'logistic': Logistic regression- works for 'rfe_coefs' or 'combinatorial' strategies
  
+            - 'reserved_inputs' (list of str): Specifies inputs that should not be eliminated
             - 'regularization' (float): Regularization coefficient for 'ridge' model
             - 'output_encoding' (str): Output encoding for 'logistic' model- either 'categorical' or 'scalar_threshold'
             - 'output_threshold' (float): Output threshold for 'logistic' model with 'scalar_threshold' encoding- values greater than or equal to the output threshold are flagged True, and the others are flagged False.
@@ -1105,6 +1093,8 @@ class Client(object):
             - 'multiclass_averaging' (str): Determines how classifier performance metrics are averaged over classes- either 'weighted', 'macro', or 'micro'
             - 'penalty' (str): Specification of penalty function for 'logistic' model- either 'none', 'l1', or 'l2'
             - 'cost' (float): Inverse regularization coefficient for 'logistic' model
+            - 'n_estimators' (int): Number of decision trees to train for random forest models
+            - 'max_depth' (int): Maximum tree depth for random forest models (optional- omit this to handle tree depth automatically) 
             - 'min_feats' (int): Minimum number of features to inspect- only for 'combinatorial' strategy
             - 'max_feats' (int): Maximum number of features to inspect- for 'combinatorial', 'rfe_coefs', and 'variances'
             - 'n_splits_cv' (int): Number of cross-validation splits- for 'rfe_coefs' or 'combinatorial'
@@ -1192,6 +1182,7 @@ class Client(object):
                 - 'none': No model is trained- the model is only for input space operations 
                 - 'ridge': Ridge regression
                 - 'logistic': Logistic regression
+                - 'rf_regressor': Random forest regression 
                 - 'gp': Gaussian Process regression- use caution with large training sets 
  
             - 'regularization' (float): Regularization coefficient for 'ridge' model
@@ -1201,6 +1192,8 @@ class Client(object):
             - 'multiclass_averaging' (str): Determines how classifier performance metrics are averaged over classes- either 'weighted', 'macro', or 'micro'
             - 'penalty' (str): Specification of penalty function- for 'logistic' model- either 'none', 'l1', or 'l2'
             - 'cost' (float): Inverse regularization coefficient for 'logistic' model
+            - 'n_estimators' (int): Number of decision trees to train for random forest models
+            - 'max_depth' (int): Maximum tree depth for random forest models (optional- omit this to handle tree depth automatically) 
             - 'kernel' (str): Kernel selection for 'gp' model- either 'rbf' or 'inv_exp' 
             - 'kernel_width' (float): Characteristic width of kernel for 'gp' model 
             - 'noise_estimate' (float): Diagonal noise estimate for 'gp' model kernel regularization 
@@ -1256,7 +1249,7 @@ class Client(object):
 
     # APPLY MODEL
 
-    def apply_models(self,model_ids,samples,run_featurization=True,run_transformation=True):
+    def apply_models(self,dataset_id,model_ids,samples,run_featurization=True,run_transformation=True):
         """Apply multiple models to one or more samples by providing model inputs.
 
         Parameters
@@ -1280,18 +1273,11 @@ class Client(object):
         """
         if isinstance(samples,pd.DataFrame):
             samples = samples.to_dict(orient='index')
-        pid_resp = self.session.get(
-            self.service_url+'process_id', 
-            data = {
-                'process_tag':'APPLY_MODELS_'+uuid.uuid4().hex
-            }
-        )
-        procid = pid_resp.json()['process_id']
         resp = self.session.post(
             self.service_url+'api/apply_models', 
             json = {
+                'dataset_id':dataset_id,
                 'model_ids':model_ids,
-                'process_id':procid,
                 'run_featurization':run_featurization,
                 'run_transformation':run_transformation,
                 'samples':samples
@@ -1349,13 +1335,13 @@ class Client(object):
         else:
             return {'success':False,'status_code':resp.status_code}
 
-    def download_apply_model_results(self,model_id,process_id,wait=False,wait_interval=1.,output_path=None):
+    def download_apply_model_results(self,dataset_id,process_id,wait=False,wait_interval=1.,output_path=None):
         """Fetch results or progress report for a model execution.
 
         Parameters
         ----------
-        model_id : int
-            Integer id of the model being applied.
+        dataset_id : int
+            Integer id of the dataset to which the model belongs.
         process_id : int
             Integer id, as assigned to the process when it was started
         wait : bool
@@ -1372,6 +1358,7 @@ class Client(object):
             Dict containing progress report, results, or error report
         """
         res = self.download_process_results(
+            dataset_id,
             process_id,
             wait=wait,
             wait_interval=wait_interval,
@@ -1382,11 +1369,13 @@ class Client(object):
         res['df'] = res.get('data')
         return res
 
-    def download_process_results(self,process_id,wait=False,wait_interval=1.,make_dataframe=False,output_path=''):
+    def download_process_results(self,dataset_id,process_id,wait=False,wait_interval=1.,make_dataframe=False,output_path=''):
         """Fetch results or progress report for any process by process id.
 
         Parameters
         ----------
+        dataset_id : int
+            Integer id, indicates the parent dataset of the process 
         process_id : int
             Integer id, as assigned to the process when it was started
         wait : bool
@@ -1426,14 +1415,18 @@ class Client(object):
         if procstat['status'] == 'FINISHED':
             res = self.session.get(
                 self.service_url+'download_file',
-                json={'file_tag':'RESULTS_'+str(process_id)}
+                json={'file_tag':'{}.RESULTS_{}'.format(dataset_id,process_id)}
             )
             data = res._content
+            df = None
             if make_dataframe:
-                data = pd.read_csv(BytesIO(res._content),index_col='id')
-                if output_path:
-                    data.to_csv(output_path)
-            return {'success':True,'data':data}
+                try:
+                    df = pd.read_csv(BytesIO(res._content))
+                    if output_path:
+                        df.to_csv(output_path)
+                except:
+                    pass 
+            return {'success':True,'data':data,'df':df}
         else:
             return procstat
 
@@ -1499,9 +1492,9 @@ class Client(object):
             The values are dicts of settings for generating candidates for that substance. 
             Each dict may have the following keys:
 
-            - 'candidates' (str): Specifies a candidate set. This should be the name of an available candidate set, or 'single' to specify a single substance.
+            - 'candidates' (str): Specifies a candidate substance set. This should be the name of an available substance set, or 'single' to specify a single substance.
             - 'selection' (str): If 'candidates'=='single', this is the name of the substance to use 
-            - 'conc_source' (str): If the substance has a concentration, this specifies how to generate scalar concentration values. Either 'single_value' or 'grid'.
+            - 'conc_source' (str): If the substance has a concentration, this specifies how to generate scalar concentration values. Either 'single_value', 'grid', or 'zip_sequence'.
             - 'conc_value' (float): If 'conc_source'=='single_value', this is the value to use
             - 'conc_minvalue' (float): If 'conc_source'=='grid', this is the minimum value.
             - 'conc_maxvalue' (float): If 'conc_source'=='grid', this is the maximum value.
@@ -1514,7 +1507,7 @@ class Client(object):
             The values are dicts of settings for generating candidate scalar values. 
             Each dict may have the following keys:
 
-            - 'source' (str): Specifies how to generate scalar values. Either 'single_value' or 'grid'.
+            - 'source' (str): Specifies how to generate scalar values. Either 'single_value', 'grid', or 'zip_sequence'.
             - 'value' (float): If 'source'=='single_value', this is the value to use
             - 'minvalue' (float): If 'source'=='grid', this is the minimum value.
             - 'maxvalue' (float): If 'source'=='grid', this is the maximum value.
@@ -1559,7 +1552,7 @@ class Client(object):
             Basic screening settings and input space handling. Keys:
 
             - 'max_n_keep' (int): Maximum number of candidates to keep. Keep as few as possible to keep the screening operation running fast and reduce garbage data accumulation. 
-            - 'joint_screening_strategy' (str): Specification of strategy for evaluating candidates jointly. Acquisition values are combined either by weighted mean or by weighted geometric mean. Options are 'mean' or 'geometric_mean'.
+            - 'joint_screening_strategy' (str): Specification of strategy for evaluating candidates jointly. Acquisition values are combined either by weighted mean or by weighted geometric mean. Options are (the default) 'mean' or 'geometric_mean'.
             - 'input_space_strategy' (str): Screening strategy for input space. Options:
 
                 - 'ignore' : Input space screening is skipped.
@@ -1601,7 +1594,7 @@ class Client(object):
         data = {'design_id':design_id,'process_id':pid_resp.json()['process_id']}
         data.update(settings)
         for nm,stgs in model_settings.items():
-             for stgnm,stgval in stgs.items():
+            for stgnm,stgval in stgs.items():
                 data[nm+'_'+stgnm] = stgval
         resp = self.session.post(
             self.service_url+'screen_candidates', 
